part of '../flutterx_live_data.dart';

/// A [Widget] whose content is built from the observed [LiveData] list
class MultipleLiveDataBuilder extends StatelessWidget {
  /// Data list which is observed
  final List<LiveData> data;

  /// Widget builder function
  final LiveWidgetBuilder<List> builder;

  /// Displayed when [data] is uninitialized, defaults to [SizedBox.shrink]
  final Widget? placeholder;

  /// Creates [builder] from [data]
  const MultipleLiveDataBuilder({
    Key? key,
    required this.data,
    required this.builder,
    this.placeholder,
  }) : super(key: key);

  /// Check [MultipleLiveDataBuilder.new]
  static MultipleLiveDataBuilder with2<X1, X2>(
          {Key? key,
          required LiveData<X1> x1,
          required LiveData<X2> x2,
          required Widget Function(BuildContext context, X1, X2) builder,
          Widget? placeholder}) =>
      MultipleLiveDataBuilder(
          data: [x1, x2],
          builder: (context, values) => builder(context, values[0], values[1]),
          placeholder: placeholder);

  /// Check [MultipleLiveDataBuilder.new]
  static MultipleLiveDataBuilder with3<X1, X2, X3>(
          {Key? key,
          required LiveData<X1> x1,
          required LiveData<X2> x2,
          required LiveData<X3> x3,
          required Widget Function(BuildContext context, X1, X2, X3) builder,
          Widget? placeholder}) =>
      MultipleLiveDataBuilder(
          data: [x1, x2, x3],
          builder: (context, values) => builder(context, values[0], values[1], values[2]),
          placeholder: placeholder);

  /// Check [MultipleLiveDataBuilder.new]
  static MultipleLiveDataBuilder with4<X1, X2, X3, X4>(
          {Key? key,
          required LiveData<X1> x1,
          required LiveData<X2> x2,
          required LiveData<X3> x3,
          required LiveData<X4> x4,
          required Widget Function(BuildContext context, X1, X2, X3, X4) builder,
          Widget? placeholder}) =>
      MultipleLiveDataBuilder(
          data: [x1, x2, x3, x4],
          builder: (context, values) => builder(context, values[0], values[1], values[2], values[3]),
          placeholder: placeholder);

  @override
  StatelessElement createElement() => _MultipleLiveDataBuilderElement(this);

  @override
  Widget build(BuildContext context) {
    final values = (context as _MultipleLiveDataBuilderElement)._values;
    return values != null ? builder(context, values) : (placeholder ?? const SizedBox.shrink());
  }

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(IterableProperty<LiveData>('data', data))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<List>>.has('builder', builder)));
}

class _MultipleLiveDataBuilderElement extends StatelessElement with _LiveDataBuilderMixin<MultipleLiveDataBuilder> {
  final List<ObserverWrapper> _observers = [];
  List? _values;

  _MultipleLiveDataBuilderElement(MultipleLiveDataBuilder widget) : super(widget);

  @override
  bool dataChanged(MultipleLiveDataBuilder oldWidget, MultipleLiveDataBuilder newWidget) =>
      !listEquals(_observers.map((observer) => observer.data).toList(growable: false), newWidget.data);

  @override
  void link(MultipleLiveDataBuilder widget) => _observeAllInternal(widget.data, onChanged,
      <V>(source, onChanged) => _observers.add(source.observeForever(onChanged, dispatcher: Dispatcher.postFrame)));

  @override
  void unlink(MultipleLiveDataBuilder widget) {
    for (final observer in _observers) observer.dispose();
    _observers.clear();
    _values = null;
  }

  @override
  void onChanged(dynamic values) {
    _values = values;
    super.onChanged(values);
  }
}
